from __future__ import absolute_import
from botocore.exceptions import ClientError
import codecs, os, sys, json
from collections import OrderedDict

import trollius as asyncio
from trollius import From, Return

from secret.vault import Kms
from secret.cli import prepare

BOTO_DEFAULT = ''
ENCODING = 'utf-8'
PY3 = (sys.version_info.major == 3)

class Storage(object):
    def list(self, **kw): pass
    def get(self, key, **kw): pass
    def put(self, key, value, **kw): pass

class S3(Storage):
    """
    A single bucket holds multiple projects.
    """
    def __init__(self, session, vault, vaultkey, region, prefix, project, env='default'):
        self.client = session.client('s3')
        self.bucket = vault
        self.vault = Kms(session=session, key=vaultkey)
        self.vaultkey = vaultkey
        self.region = region
        self.prefix = prefix
        self.project = project
        self.env = env + '/'

        if self.prefix:
            self.prefix += '/'

    @asyncio.coroutine
    def create_bucket(self, name=None, **kw):
        # administrator creates bucket for projects
        name = name or self.bucket
        result = self.client.create_bucket(Bucket=name)
        raise Return(result)

    @asyncio.coroutine
    def setup(self, vault, project, **kw):
        # project-user creates a folder inside bucket
        if not all([vault, project]):
            sys.exit("Error! Vault and/or project undefined")
        project_s3 = project.rstrip('/')+'/'
        try:
            result = yield From(self.get(project_s3, in_group_check=False))
            res = "Error! A project with this name already exists"
        except ClientError as e:
            result = yield From(self.put(project_s3, '', prefixify=False))
            self.project.save(dict(vault=vault, project=project, key=self.vaultkey, region=self.region))
            res = "Success! Project created. Configuration stored in %s"%self.project.name
        raise Return(res)

    @asyncio.coroutine
    def list_backend(self, prefix=None, **kw):
        prefix = prefix or self.prefixify('')
        raise Return(self.client.list_objects(Bucket=self.bucket, Prefix=prefix, MaxKeys=1000, Delimiter=kw.get('delimiter', BOTO_DEFAULT)))

    @asyncio.coroutine
    def list(self, key=None, **kw):
        key = self.prefixify(key)
        response = yield From(self.list_backend(prefix=key, **kw))
        raise Return([self.prefixify(k['Key'], reverse=True) for k in response.get('Contents', [])])

    @asyncio.coroutine
    def config(self, key=None, **kw):
        args = prepare()
        contents = yield From(self.list_backend(prefix=self.prefixify(key)))
        contents = contents.get('Contents', [])
        kw.setdefault('in_group_check', False)
        kw.setdefault('no_files', True)
        tasks = [self._get(key=obj['Key'], **kw) for obj in contents]
        try:
            results = yield From(asyncio.gather(*tasks))
        except ValueError as e:
            raise Return({})
        result = OrderedDict()
        project_env = "{}/{}".format(args.project, args.env)
        for k, v in enumerate(results):
            if kw.get('in_group_check'): # for grouped keys strip prefixes
                key_ = contents[k]['Key'].split('/')[-1]
            else: # otherwise full key
                key_ = contents[k]['Key'].replace(project_env, '').lstrip('/')
            if args.skip_files and v['is_file']:
                continue
            result[key_] = v['data']
        raise Return(result)

    @asyncio.coroutine
    def put_backend(self, **kw):
        raise Return(self.client.put_object(**kw))

    @asyncio.coroutine
    def put(self, key, value, **kw):
        if value is None:
            raise Return("Error! No value provided.")

        if kw.get('prefixify', True):
            key = self.prefixify(key)
        is_file = False
        is_binary = False # TODO: --binary
        mode = 'rb'

        # TODO --encode flag to encode value
        if os.path.isfile(value):
            kwargs = dict(mode=mode)
            if PY3:
                kwargs = dict(mode=mode, encoding=ENCODING)
            value = codecs.open(os.path.expandvars(os.path.expanduser(value)), **kwargs).read().rstrip('\n')
            is_file = True

        if PY3:
            value = value.encode(ENCODING)

        # boto expects bytestrings
        data = self.vault.encrypt(value, is_file=is_file, is_binary=is_binary)
        data['name'] = key
        result = yield From(self.put_backend(Bucket=self.bucket, Key=key, Body=json.dumps(data)))
        if result.get('ResponseMetadata').get('HTTPStatusCode') == 200:
            raise Return("Success! Wrote: %s"%key)
        else:
            raise Return(result)

    @asyncio.coroutine
    def get_backend(self, **kw):
        raise Return(self.client.get_object(**kw))

    @asyncio.coroutine
    def _get(self, key, **kw):
        if key is None:
            raise Return("Error! No key provided.")

        key = self.prefixify(key)
        extra = {}
        if kw.get('version'):
            extra['VersionId'] = kw.get('version')
        result = yield From(self.get_backend(Bucket=self.bucket, Key=key, **extra))
        body = json.loads(result['Body'].read().decode(ENCODING))
        data = self.vault.decrypt(body)
        is_binary = body.get('is_binary', False)
        if not is_binary:
            data = data.decode(ENCODING)

        if kw.get('no_files') and body.get('is_file'):
            data = "<FILE>"
        body["data"] = data
        raise Return(body)

    @asyncio.coroutine
    def get(self, key, **kw):
        try:
            data = yield From(self._get(key, **kw))
        except Exception as e:
            if kw.get('in_group_check', True):
                # check for grouped keys
                key = key.rstrip('/')+'/'
                kw['in_group_check'] = True
                result = yield From(self.config(key=key, **kw))
                if not result:
                    result = "Error! The specified key does not exist."
                raise Return(result)
            raise
        raise Return(data["data"])

    @asyncio.coroutine
    def delete_backend(self, **kw):
        raise Return(self.client.delete_object(**kw))

    @asyncio.coroutine
    def delete(self, key, **kw):
        key = self.prefixify(key)
        result = yield From(self.delete_backend(Bucket=self.bucket, Key=key))
        if result.get('ResponseMetadata').get('HTTPStatusCode') == 204:
            raise Return("Success! Deleted: %s"%key)
        else:
            raise Return(result)

    @asyncio.coroutine
    def versions(self, **kw):
        prefix = kw.get('prefix', self.prefix)
        response = self.client.list_object_versions(Bucket=self.bucket, Prefix=self.prefix)
        versions = []
        for k in response.get('Versions', []):
            key = self.prefixify(k['Key'], reverse=True)
            if kw.get('key') and kw.get('key')!=key: continue
            versions.append(dict(key=key, version=k['VersionId'], is_latest=k['IsLatest'], modified=k['LastModified']))
        raise Return(versions)

    @asyncio.coroutine
    def envs(self, **kw):
        contents = yield From(self.list_backend(prefix=self.prefix, delimiter='/'))
        raise Return(list(k['Prefix'].split('/')[1] for k in contents.get('CommonPrefixes', [])))

    def prefixify(self, key, reverse=False):
        if reverse:
            if self.prefix and self.env:
                key = key.replace('{}{}'.format(self.prefix, self.env), '')
        elif key is not None:
            if self.env:
                if self.prefix not in key and self.env not in key:
                    key = '{}{}{}'.format(self.prefix, self.env, key)
            else:
                if self.prefix not in key:
                    key = '{}{}'.format(self.prefix, key)
        key = key or ''
        key = key.replace('//','/')
        return key



from Crypto.Cipher import AES
from Crypto.Hash.HMAC import HMAC
from Crypto.Hash import SHA256
from Crypto.Random import get_random_bytes

class Cipher:
    def encrypt(self, value, **kw): pass
    def decrypt(self, value, **kw): pass

class AesEax(Cipher):
    def __init__(self, key):
        self.key = key

    def cipher(self, nonce):
        return AES.new(self.key, AES.MODE_EAX, nonce=nonce)

    @classmethod
    def hmac(self, key, msg):
        return HMAC(key, msg=msg, digestmod=SHA256).hexdigest()

    def encrypt(self, value):
        nonce = get_random_bytes(16)
        c_text, tag = self.cipher(nonce).encrypt_and_digest(value)
        return c_text, nonce, tag

    def decrypt(self, value, nonce, tag):
        return self.cipher(nonce).decrypt_and_verify(value, tag)



from __future__ import absolute_import

try:# PY2
    import ConfigParser as configparser
except:
    import configparser
import argparse, sys, os
from secret.project import get_project
from secret.templates import statusT, helpT

DATAFILE = '.secret'
ALIASES = {'ls':'list','rm':'delete','set':'put',}

def print_status(args):
    aws_profile = bool(os.getenv("AWS_PROFILE"))
    aws_key = bool(os.getenv('AWS_ACCESS_KEY_ID'))
    aws_secret = bool(os.getenv('AWS_SECRET_ACCESS_KEY'))
    print(statusT.format(args.vault, args.vaultkey, args.project, args.env, aws_profile, aws_key, aws_secret))

def prepare():
    p = argparse.ArgumentParser()
    p.add_argument("action", help="list,get,put,envs,config,setup")
    p.add_argument("key", nargs="?", default=None)
    p.add_argument("value", nargs="?", default=None)
    p.add_argument("--version", default=None)
    p.add_argument("--region", help="AWS region of KMS and S3", default=None)
    p.add_argument("--vaultkey", help="Name of KMS key", default=None)
    p.add_argument("--vault", help="Name of vault (eg. S3 bucket)", default=None)
    p.add_argument("-P", "--project", help="Name of project (eg. S3 'folder')", default=None)
    p.add_argument("--env", help="Environment namespace for keys", default='default')
    p.add_argument("-F","--fmt", help="Output format", default='console')
    p.add_argument("-o","--output", help="Output to console (default) or given filepath", default='')
    p.add_argument("--datafile", default=DATAFILE)
    p.add_argument("--debug", default=None)
    p.add_argument("--skip-files", action='store_true', default=False)
    if len(sys.argv) == 1:
        sys.exit(p.print_help())
    args = p.parse_args()
    args.action = ALIASES.get(args.action, args.action)

    project = get_project(args.datafile)

    # Arguments preference: CLI -> .secret -> globals

    args.project = args.project if (args.project is not None) else project.load().get('project', '')
    args.vault = args.vault or project.load().get('vault')
    args.vaultkey = args.vaultkey or project.load().get('key')
    args.region = args.region or project.load().get('region')

    secret_profile = os.getenv("SECRET_PROFILE", "default")
    config = configparser.SafeConfigParser()
    config.read(os.path.expanduser('~/.secret/credentials'))
    if config.has_section(secret_profile):
        if not args.vault:
            args.vault = config.get(secret_profile, 'vault', raw=0)
        if not args.vaultkey:
            args.vaultkey = config.get(secret_profile, 'vaultkey', raw=0)
        if not args.region:
            args.region = config.get(secret_profile, 'region', raw=0)

    if not all([args.vault]):
        sys.exit("Vault configuration undefined: --vault, --vaultkey")

    if args.action == 'help':
        print_status(args)
        sys.exit(helpT)

    return args



#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import print_function
import sys, six, json
from tabulate import tabulate

ENCODING = 'utf-8'
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
PY2 = (sys.version_info.major == 2)

class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        sys.stdout = self._stdout

def fmt_console(result):
    def is_str(result):
        try:
            return isinstance(result, basestring)
        except NameError:
            return isinstance(result, str)
    if any(isinstance(result, k) for k in [list]):
        print("Keys:")
        for k in result:
            print(k)
    elif isinstance(result, bytes):
        print(result)
    elif is_str(result):
        print(result)
    else:
        table = [["Key", "Value"]]
        for k,v in six.iteritems(result):
            table.append([k,v])
        print(tabulate(table, numalign='left', tablefmt='plain'))

def fmt_docker(result):
    c = []
    for k,v in six.iteritems(result):
        c.append('-e '+k+'="'+v+'"')
    print(" ".join(c), end='')

def fmt_json(result):
    print(json.dumps(result), end='')

def has_error(result):
    try:
        if result.startswith('Error'):
            return True
    except Exception:
        return False

def prettyprint(result, args):
    with Capturing() as output:
        if args.fmt == 'docker':
            fmt_docker(result)
        elif args.fmt == 'json':
            fmt_json(result)
        else:
            fmt_console(result)

    pretty = "\n".join(output)
    if PY2:
        pretty = pretty.encode(ENCODING)
    if args.output == '':
        print(pretty)
    else:
        if not has_error(result):
            with open(args.output, 'w') as f:
                f.write(pretty)

    if has_error(result):
        sys.exit(1)




statusT = """
VAULT: {}
KEY: {}
PROJECT: {} 
ENV: {}
AWS/profile? {}
AWS/key? {}
AWS/secret? {}"""

helpT = """
1. Run setup
2. Check that AWS environment variables for profile OR key+secret are set
"""
